@@ -1,6 +1,6 @@
 ###############################################################################
 #
-# Copyright (c) 2009 Cedric Krier.
+# Copyright (c) 2009-2017 Cedric Krier.
 # Copyright (c) 2007, 2008 OpenHex SPRL. (http://openhex.com) All Rights
 # Reserved.
 #
@@ -19,8 +19,6 @@
 #
 ###############################################################################
 
-__metaclass__ = type
-
 import re
 try:
     # requires python 2.5+
@@ -28,15 +26,18 @@ try:
 except ImportError:
     from md5 import md5
 
+import base64
+import mimetypes
 import time
 import urllib
 import zipfile
-from cStringIO import StringIO
+from io import BytesIO
 from copy import deepcopy
+import datetime
+from decimal import Decimal
 
 
 import warnings
-warnings.filterwarnings('always', module='relatorio.templates.opendocument')
 
 import lxml.etree
 import genshi
@@ -47,7 +48,7 @@ from genshi.filters.transform import ENTER, EXIT
 from genshi.core import Stream
 from genshi.template.interpolation import PREFIX
 
-
+import relatorio
 from relatorio.templates.base import RelatorioStream
 from relatorio.reporting import Report, MIMETemplateLoader
 try:
@@ -55,28 +56,36 @@ try:
 except ImportError:
     ChartTemplate = type(None)
 
+__metaclass__ = type
+warnings.filterwarnings('always', module='relatorio.templates.opendocument')
+
 GENSHI_EXPR = re.compile(r'''
         (/)?                                 # is this a closing tag?
-        (for|if|choose|when|otherwise|with)  # tag directive
+        (for|if|choose|when|otherwise|with|
+         attrs|content|replace|strip)        # tag directive
         \s*
-        (?:\s(\w+)=["'](.*)["']|$)           # match a single attr & its value
+        (?:\s([\w:-]+)=["'](.*)["']|$)       # match a single attr & its value
         |
         .*                                   # or anything else
         ''', re.VERBOSE)
+GENSHI_CLOSING_DIRECTIVE = ['for',
+                            'if',
+                            'choose',
+                            'when',
+                            'otherwise',
+                            'with',
+                            ]
 
-EXTENSIONS = {'image/png': 'png',
-              'image/jpeg': 'jpg',
-              'image/bmp': 'bmp',
-              'image/gif': 'gif',
-              'image/tiff': 'tif',
-              'image/xbm': 'xbm',
-             }
-
-RELATORIO_URI = 'http://relatorio.openhex.org/'
+RELATORIO_URI = 'relatorio'
 GENSHI_URI = 'http://genshi.edgewall.org/'
 MANIFEST = 'META-INF/manifest.xml'
+META = 'meta.xml'
+THUMBNAILS = 'Thumbnails'
 output_encode = genshi.output.encode
 EtreeElement = lxml.etree.Element
+XML_INVALID_CHAR_EXPR = re.compile(
+    # from https://www.w3.org/TR/REC-xml/#charsets
+    u'[\x00-\x08\x0b\x0c\x0e-\x1F\uD800-\uDFFF\uFFFE\uFFFF]')
 
 # A note regarding OpenDocument namespaces:
 #
@@ -97,11 +106,6 @@ EtreeElement = lxml.etree.Element
 # parent node defines it. Unfortunately, lxml doesn't support this:
 # the nsmap attribute of Element objects is (currently) readonly.
 
-def guess_type(val):
-    if isinstance(val, (str, unicode)):
-        return 'string'
-    elif isinstance(val, (int, float, long)):
-        return 'float'
 
 class OOTemplateError(genshi.template.base.TemplateSyntaxError):
     "Error to raise when there is a SyntaxError in the genshi template"
@@ -121,10 +125,13 @@ class ImageHref:
             bitstream = bitstream(**self.context).render()
         elif isinstance(bitstream, ChartTemplate):
             bitstream = bitstream.generate(**self.context).render()
+        elif not hasattr(bitstream, 'seek') or not hasattr(bitstream, 'read'):
+            bitstream = BytesIO(bitstream)
         bitstream.seek(0)
         file_content = bitstream.read()
         name = md5(file_content).hexdigest()
-        path = 'Pictures/%s.%s' % (name, EXTENSIONS[mimetype])
+        path = 'Pictures/%s%s' % (
+            name, mimetypes.guess_extension(mimetype))
         if path not in self.zip.namelist():
             self.zip.writestr(path, file_content)
             self.manifest.add_file_entry(path, mimetype)
@@ -140,8 +147,9 @@ class ImageDimension:
     def __call__(self, expr, width, height):
         # expr could be (bitstream, mimetype)
         # or (bitstreamm mimetype, width, height)
-        if len(expr) == 4:
-            width, height = expr[2:]
+        if len(expr) >= 4:
+            width, height = (
+                i or j for i, j in zip(expr[2:4], [width, height]))
         attrs = {}
         if width:
             attrs['{%s}width' % self.namespaces['svg']] = width
@@ -169,6 +177,19 @@ class ColumnCounter:
                                         self.counters.get(table_name, 0))
 
 
+class ExpressionCache:
+    """A class used to cache result of expression evaluation"""
+    def __init__(self):
+        self.cache = {}
+
+    def store(self, expression_id, expression):
+        self.cache[expression_id] = expression
+        return expression
+
+    def get(self, expression_id):
+        return self.cache[expression_id]
+
+
 def wrap_nodes_between(first, last, new_parent):
     """An helper function to move all nodes between two nodes to a new node
     and add that new node to their former parent. The boundary nodes are
@@ -189,16 +210,37 @@ def wrap_nodes_between(first, last, new_parent):
         # remove it from its previous parent
         new_parent.append(node)
     old_parent.replace(first, new_parent)
+    new_parent.tail = last.tail
     old_parent.remove(last)
 
 
+def remove_node_keeping_tail(node):
+    """Remove the node from the tree but keeping tail by appending to the
+    previous or parent node.
+    """
+    parent = node.getparent()
+    if node.tail:
+        previous = node.getprevious()
+        if previous is not None:
+            if not previous.tail:
+                previous.tail = node.tail
+            else:
+                previous.tail += node.tail
+        else:
+            if not parent.text:
+                parent.text = node.tail
+            else:
+                parent.text += node.tail
+    parent.remove(node)
+
+
 def update_py_attrs(node, value):
     """An helper function to update py_attrs of a node.
     """
     if not value:
         return
     py_attrs_attr = '{%s}attrs' % GENSHI_URI
-    if not py_attrs_attr in node.attrib:
+    if py_attrs_attr not in node.attrib:
         node.attrib[py_attrs_attr] = value
     else:
         node.attrib[py_attrs_attr] = \
@@ -206,6 +248,14 @@ def update_py_attrs(node, value):
                 (node.attrib[py_attrs_attr], value)
 
 
+def escape_xml_invalid_chars(value, repl=' '):
+    "Replace invalid characters for XML."
+    if isinstance(value, basestring):
+        return XML_INVALID_CHAR_EXPR.sub(repl, value)
+    else:
+        return value
+
+
 class Template(MarkupTemplate):
 
     def __init__(self, source, filepath=None, filename=None, loader=None,
@@ -213,6 +263,7 @@ class Template(MarkupTemplate):
         self.namespaces = {}
         self.inner_docs = []
         self.has_col_loop = False
+        self._source = None
         super(Template, self).__init__(source, filepath, filename, loader,
                                        encoding, lookup, allow_exec)
 
@@ -221,7 +272,21 @@ class Template(MarkupTemplate):
 
         It adds genshi directives and finds the inner docs.
         """
-        zf = zipfile.ZipFile(self.filepath)
+        if not self.filepath:
+            if hasattr(source, 'read') and hasattr(source, 'mode'):
+                if 'U' in source.mode:
+                    # TemplateLoader of Genshi <= 0.6 open files with universal
+                    # newlines which is not suitable for zipfile
+                    raise ValueError('filepath is required '
+                        'if source is openned with universal newlines')
+                else:
+                    # source could be closed before generate calls
+                    source = BytesIO(source.read())
+        else:
+            source = self.filepath
+        self._source = source
+        self.filepath = None  # Prevent zip content in traceback
+        zf = get_zip_file(source)
         content = zf.read('content.xml')
         styles = zf.read('styles.xml')
 
@@ -243,8 +308,8 @@ class Template(MarkupTemplate):
                                        encoding)
             content_files.append((c_path, c_parsed))
             styles_files.append((s_path, s_parsed))
-
         zf.close()
+
         parsed = []
         for fpath, fparsed in content_files + styles_files:
             parsed.append((genshi.core.PI, ('relatorio', fpath), None))
@@ -255,7 +320,7 @@ class Template(MarkupTemplate):
     def insert_directives(self, content):
         """adds the genshi directives, handle the images and the innerdocs.
         """
-        tree = lxml.etree.parse(StringIO(content))
+        tree = lxml.etree.parse(BytesIO(content))
         root = tree.getroot()
 
         # assign default/fake namespaces so that documents do not need to
@@ -277,12 +342,24 @@ class Template(MarkupTemplate):
         self.namespaces['py'] = GENSHI_URI
         self.namespaces['relatorio'] = RELATORIO_URI
 
+        self._remove_soft_page_break(tree)
         self._invert_style(tree)
         self._handle_relatorio_tags(tree)
         self._handle_images(tree)
         self._handle_innerdocs(tree)
         self._escape_values(tree)
-        return StringIO(lxml.etree.tostring(tree))
+        return BytesIO(lxml.etree.tostring(tree))
+
+    def _remove_soft_page_break(self, tree):
+        "remove soft-page-break tag and use-soft-page-break attribute"
+        xpath_expr = "//text:soft-page-break"
+        for node in tree.xpath(xpath_expr, namespaces=self.namespaces):
+            remove_node_keeping_tail(node)
+
+        xpath_expr = "//office:text[@text:use-soft-page-breaks]"
+        text_namespace = self.namespaces['text']
+        for node in tree.xpath(xpath_expr, namespaces=self.namespaces):
+            node.attrib.pop('{%s}use-soft-page-breaks' % text_namespace)
 
     def _invert_style(self, tree):
         "inverts the text:a and text:span"
@@ -322,7 +399,7 @@ class Template(MarkupTemplate):
                 raise OOTemplateError("No expression in the tag",
                                       self.filepath)
             closing, directive, attr, attr_val = \
-                    GENSHI_EXPR.match(expr).groups()
+                GENSHI_EXPR.match(expr).groups()
             is_opening = closing != '/'
 
             warn_msg = None
@@ -339,12 +416,17 @@ class Template(MarkupTemplate):
                                 % opened_tags[-1].text
                 warnings.warn(warn_msg)
 
-            if directive is not None:
+            if directive in GENSHI_CLOSING_DIRECTIVE:
                 # map closing tags with their opening tag
                 if is_opening:
-                    opened_tags.append(statement)
+                    opened_tags.append((statement, directive))
                 else:
-                    closing_tags[id(opened_tags.pop())] = statement
+                    opening_statement, opening_directive = opened_tags.pop()
+                    assert directive == opening_directive, (
+                        "Wrong pairing tags between '%s' and '%s'"
+                        % (opening_statement.text.encode('utf-8'),
+                           statement.text.encode('utf-8')))
+                    closing_tags[id(opening_statement)] = statement
             # - we only need to return opening statements
             if is_opening:
                 r_statements.append((statement,
@@ -362,11 +444,17 @@ class Template(MarkupTemplate):
         table_namespace = self.namespaces['table']
         table_row_tag = '{%s}table-row' % table_namespace
         table_cell_tag = '{%s}table-cell' % table_namespace
+        text_namespace = self.namespaces['text']
+        text_style_attributes = [s % text_namespace for s in [
+                '{%s}class-names', '{%s}cond-style-name', '{%s}style-name']]
 
-        office_name = '{%s}value' % self.namespaces['office']
+        office_value = '{%s}value' % self.namespaces['office']
         office_valuetype = '{%s}value-type' % self.namespaces['office']
+        if 'calcext' in self.namespaces:
+            calcext_valuetype = '{%s}value-type' % self.namespaces['calcext']
+        else:
+            calcext_valuetype = None
 
-        py_attrs_attr = '{%s}attrs' % GENSHI_URI
         py_replace = '{%s}replace' % GENSHI_URI
 
         r_statements, closing_tags = self._relatorio_statements(tree)
@@ -375,7 +463,7 @@ class Template(MarkupTemplate):
             expr, directive, attr, a_val = parsed
 
             # If the node is a genshi directive statement:
-            if directive is not None:
+            if directive in GENSHI_CLOSING_DIRECTIVE:
                 opening = r_node
                 closing = closing_tags[id(r_node)]
 
@@ -398,7 +486,7 @@ class Template(MarkupTemplate):
                         pass
                     o_ancestors.append(node)
                 assert ancestor is not None, \
-                       "No common ancestor found for opening and closing tag"
+                    "No common ancestor found for opening and closing tag"
 
                 outermost_o_ancestor = o_ancestors[-1]
                 outermost_c_ancestor = c_ancestors[-1]
@@ -427,20 +515,50 @@ class Template(MarkupTemplate):
                 # - we delete the closing statement (and its ancestors)
                 wrap_nodes_between(outermost_o_ancestor, outermost_c_ancestor,
                                    genshi_node)
+            elif directive:
+                # find the first parent with the same tag name as the attribute
+                parent = r_node
+                namespace, name = attr.split(':')
+                attr = '{%s}%s' % (self.namespaces[namespace], name)
+                while parent is not None and parent.tag != attr:
+                    parent = parent.getparent()
+                assert parent is not None, "Parent not found"
+
+                # add the py:attribute to the parent
+                py_attr = '{%s}%s' % (GENSHI_URI, directive)
+                parent.attrib[py_attr] = a_val
+
+                # remove the directive node
+                r_node.getparent().remove(r_node)
             else:
+                def has_style(node):
+                    return any(attr in node.attrib
+                               for attr in text_style_attributes)
                 # It's not a genshi statement it's a python expression
-                r_node.attrib[py_replace] = expr
-                parent = r_node.getparent().getparent()
-                if parent is None or parent.tag != table_cell_tag:
+                parent = r_node.getparent()
+                grand_parent = parent.getparent()
+                # Guess type only if it is the only value in the cell
+                # and its parent has no style
+                if (grand_parent is None
+                        or grand_parent.tag != table_cell_tag
+                        ) or len(grand_parent) != 1 or has_style(parent):
+                    r_node.attrib[py_replace] = (
+                        '__relatorio_escape_invalid_chars(%s)' % expr)
                     continue
 
+                cache_id = id(r_node)
+                r_node.attrib[py_replace] = ("__relatorio_get_cache(%s)" %
+                                             cache_id)
                 # The grand-parent tag is a table cell we should set the
                 # correct value and type for this cell.
-                dico = "{'%s': %s, '%s': __relatorio_guess_type(%s)}"
-                update_py_attrs(parent, dico %
-                        (office_name, expr, office_valuetype, expr))
-                parent.attrib.pop(office_valuetype, None)
-                parent.attrib.pop(office_name, None)
+                dico = ('__relatorio_guess_type('
+                        '__relatorio_store_cache(%s, %s))')
+                update_py_attrs(grand_parent, dico % (cache_id, expr))
+                for attr in [office_value,
+                             office_valuetype,
+                             calcext_valuetype]:
+                    if attr:
+                        grand_parent.attrib.pop(attr, None)
 
     def _handle_column_loops(self, statement, ancestor, opening,
                              outer_o_node, outer_c_node):
@@ -452,7 +570,6 @@ class Template(MarkupTemplate):
         table_col_tag = '{%s}table-column' % table_namespace
         table_num_col_attr = '{%s}number-columns-repeated' % table_namespace
 
-        py_attrs_attr = '{%s}attrs' % GENSHI_URI
         repeat_tag = '{%s}repeat' % RELATORIO_URI
 
         # table node (it is not necessarily the direct parent of ancestor)
@@ -534,12 +651,8 @@ class Template(MarkupTemplate):
             for tag in to_split:
                 tag_pos = table_node.index(tag)
                 num = int(tag.attrib.pop(table_num_col_attr))
-                new_tags = [deepcopy(tag) for _ in range(num)]
-                table_node[tag_pos:tag_pos+1] = new_tags
-
-            # recompute the list of column headers as it could
-            # have changed.
-            coldefs = list(table_node.iterchildren(table_col_tag))
+                new_tags = [deepcopy(tag) for i in range(num)]
+                table_node[tag_pos:tag_pos + 1] = new_tags
 
             # compute the column header nodes corresponding to
             # the opening and closing tags.
@@ -549,10 +662,10 @@ class Template(MarkupTemplate):
             # add a <relatorio:repeat> node around the column
             # definitions nodes
             attribs = {
-               "opening": str(opening_pos),
-               "closing": str(closing_pos),
-               "table": table_name
-            }
+                "opening": str(opening_pos),
+                "closing": str(closing_pos),
+                "table": table_name,
+                }
             repeat_node = EtreeElement(repeat_tag, attrib=attribs,
                                        nsmap={'relatorio': RELATORIO_URI})
             wrap_nodes_between(first, last, repeat_node)
@@ -599,7 +712,7 @@ class Template(MarkupTemplate):
         rows_to_wrap = [row_node]
         assert row_node.tag == table_row_tag
         next_rows = row_node.itersiblings(table_row_tag)
-        for row_idx in range(rows_spanned-1):
+        for row_idx in range(rows_spanned - 1):
             next_row_node = next_rows.next()
             rows_to_wrap.append(next_row_node)
             # compute the start and end nodes
@@ -626,23 +739,32 @@ class Template(MarkupTemplate):
         draw_name = '{%s}name' % draw_namespace
         draw_image = '{%s}image' % draw_namespace
         py_attrs = '{%s}attrs' % self.namespaces['py']
+        end_cell_address = '{%s}end-cell-address' % self.namespaces['table']
         svg_namespace = self.namespaces['svg']
         svg_width = '{%s}width' % svg_namespace
         svg_height = '{%s}height' % svg_namespace
         xpath_expr = "//draw:frame[starts-with(@draw:name, 'image:')]"
         for draw in tree.xpath(xpath_expr, namespaces=self.namespaces):
+            cache_id = id(draw)
             d_name = draw.attrib[draw_name][6:].strip()
-            attr_expr = "__relatorio_make_href(%s)" % d_name
+            attr_expr = ("__relatorio_make_href(__relatorio_get_cache(%s))" %
+                         cache_id)
             image_node = EtreeElement(draw_image,
                                       attrib={py_attrs: attr_expr},
                                       nsmap={'draw': draw_namespace,
                                              'py': GENSHI_URI})
             draw.replace(draw[0], image_node)
-            width = draw.attrib.pop(svg_width, None)
-            height = draw.attrib.pop(svg_height, None)
-            attr_expr = "__relatorio_make_dimension(%s, '%s', '%s')" % \
-                    (d_name, width, height)
+            width = draw.attrib.pop(svg_width, '')
+            height = draw.attrib.pop(svg_height, '')
+            attr_expr = ("__relatorio_make_dimension("
+                         "__relatorio_store_cache(%s, %s), '%s', '%s')" %
+                         (cache_id, d_name, width, height))
             draw.attrib[py_attrs] = attr_expr
+            draw.attrib.pop(draw_name)
+            dico = "{'%s': (__relatorio_get_cache(%s)[4:5] or [''])[0]}"
+            update_py_attrs(draw, dico % (draw_name, cache_id))
+            # remove end-cell-address as the address specified could be wrong
+            draw.attrib.pop(end_cell_address, '')
 
     def _handle_innerdocs(self, tree):
         "finds inner_docs and adds them to the processing stack."
@@ -662,20 +784,61 @@ class Template(MarkupTemplate):
             if element.text:
                 element.text = element.text.replace(PREFIX, PREFIX * 2)
 
+    def _guess_type(self, val):
+        office_namespace = self.namespaces['office']
+        types = {'boolean': '{%s}boolean-value' % office_namespace,
+                 'currency': '{%s}currency' % office_namespace,
+                 'date': '{%s}date-value' % office_namespace,
+                 'float': '{%s}value' % office_namespace,
+                 'percentage': '{%s}value' % office_namespace,
+                 'string': '{%s}string-value' % office_namespace,
+                 'time': '{%s}time-value' % office_namespace,
+                 'void': '{%s}value' % office_namespace,
+                 }
+        attrs = dict.fromkeys(types)
+        # Missing base type for currency and percentage
+        if isinstance(val, bool):
+            type_ = 'boolean'
+            val = str(val).lower()
+        elif isinstance(val, datetime.date):
+            type_ = 'date'
+            val = val.isoformat()
+        elif isinstance(val, (int, float, long, Decimal)):
+            type_ = 'float'
+        elif isinstance(val, basestring):
+            type_ = 'string'
+            val = escape_xml_invalid_chars(val)
+        elif isinstance(val, datetime.timedelta):
+            type_ = 'time'
+            val = 'P%sD%sS' % (val.days, val.seconds)
+        else:
+            type_ = 'void'
+            val = None
+        attrs[types[type_]] = val
+        attrs['{%s}value-type' % office_namespace] = type_
+        if 'calcext' in self.namespaces:
+            attrs['{%s}value-type' % self.namespaces['calcext']] = type_
+        return attrs
+
     def generate(self, *args, **kwargs):
         "creates the RelatorioStream."
-        serializer = OOSerializer(self.filepath)
+        serializer = OOSerializer(self._source)
         kwargs['__relatorio_make_href'] = ImageHref(serializer.outzip,
                                                     serializer.manifest,
                                                     kwargs)
         kwargs['__relatorio_make_dimension'] = ImageDimension(self.namespaces)
-        kwargs['__relatorio_guess_type'] = guess_type
+        kwargs['__relatorio_guess_type'] = self._guess_type
+        kwargs['__relatorio_escape_invalid_chars'] = escape_xml_invalid_chars
 
         counter = ColumnCounter()
         kwargs['__relatorio_reset_col_count'] = counter.reset
         kwargs['__relatorio_inc_col_count'] = counter.inc
         kwargs['__relatorio_store_col_count'] = counter.store
 
+        cache = ExpressionCache()
+        kwargs['__relatorio_store_cache'] = cache.store
+        kwargs['__relatorio_get_cache'] = cache.get
+
         stream = super(Template, self).generate(*args, **kwargs)
         if self.has_col_loop:
             # Note that we can't simply add a "number-columns-repeated"
@@ -717,16 +880,102 @@ class DuplicateColumnHeaders(object):
                 yield mark, (kind, data, pos)
 
 
+def get_zip_file(source):
+    try:
+        return zipfile.ZipFile(source)
+    except zipfile.BadZipfile:
+        # ZipFile modify the position
+        if hasattr(source, 'seek'):
+            source.seek(0)
+        return zipfile.ZipFile(fod2od(source))
+
+
+def fod2od(source):
+    "Convert Flat OpenDocument to OpenDocument"
+    odt_io = BytesIO()
+    odt_zip = zipfile.ZipFile(
+        odt_io, mode='w', compression=zipfile.ZIP_DEFLATED)
+    fodt_tree = lxml.etree.parse(source)
+    fodt_root = fodt_tree.getroot()
+    office_ns = fodt_root.nsmap['office']
+    tag2files = {
+        '{%s}meta' % office_ns: ['meta'],
+        '{%s}settings' % office_ns: ['settings'],
+        '{%s}scripts' % office_ns: ['content'],
+        '{%s}font-face-decls' % office_ns: ['content', 'styles'],
+        '{%s}styles' % office_ns: ['styles'],
+        '{%s}automatic-styles' % office_ns: ['content', 'styles'],
+        '{%s}master-styles' % office_ns: ['styles'],
+        '{%s}body' % office_ns: ['content'],
+        }
+    mimetype = fodt_root.attrib['{%s}mimetype' % office_ns]
+    # mimetype should be written first to let detection through 'magic number'
+    odt_zip.writestr('mimetype', mimetype, zipfile.ZIP_STORED)
+    documents = {}
+    images = []
+    for child in fodt_root:
+        for fname in tag2files[child.tag]:
+            document = documents.get(fname)
+            if document is None:
+                document = lxml.etree.Element(
+                    '{%s}document-%s' % (office_ns, fname),
+                    nsmap=fodt_root.nsmap)
+                documents[fname] = document
+            child = deepcopy(child)
+            images.extend(extract_images(
+                    child, fodt_root.nsmap, start=len(images)))
+            document.append(child)
+    manifest = Manifest(b'''<?xml version="1.0" encoding="UTF-8"?>
+        <manifest:manifest
+        xmlns:manifest="urn:oasis:names:tc:opendocument:xmlns:manifest:1.0"/>
+        ''')
+    manifest.add_file_entry('/', mimetype)
+    for fname, document in documents.iteritems():
+        document_string = lxml.etree.tostring(document, encoding='UTF-8',
+                                              xml_declaration=True)
+        odt_zip.writestr('%s.xml' % fname, document_string)
+        manifest.add_file_entry('%s.xml' % fname, 'text/xml')
+    for fname, data, mime_type in images:
+        odt_zip.writestr(fname, data)
+        manifest.add_file_entry(fname, mime_type)
+    odt_zip.writestr(MANIFEST, str(manifest))
+    odt_zip.close()
+    return odt_io
+
+
+def extract_images(child, namespaces, start=0):
+    "Extract draw:image with binary-data and replace by href"
+    import magic
+    images = []
+    for i, image in enumerate(
+            child.xpath('//draw:image', namespaces=namespaces), start):
+        binary_data, = image.xpath(
+            './office:binary-data', namespaces=namespaces)
+        data = base64.b64decode(binary_data.text)
+        mime_type = magic.from_buffer(data, mime=True)
+        name = 'Pictures/image%s%s' % (
+            i, mimetypes.guess_extension(mime_type))
+        image.remove(binary_data)
+        xlink_ns = namespaces['xlink']
+        image.attrib['{%s}href' % xlink_ns] = name
+        images.append((name, data, mime_type))
+    return images
+
+
 class Manifest(object):
 
     def __init__(self, content):
-        self.tree = lxml.etree.parse(StringIO(content))
+        self.tree = lxml.etree.parse(BytesIO(content))
         self.root = self.tree.getroot()
         self.namespaces = self.root.nsmap
 
     def __str__(self):
-        return lxml.etree.tostring(self.tree, encoding='UTF-8',
-                                   xml_declaration=True)
+        val = lxml.etree.tostring(self.tree, encoding='UTF-8',
+                                  xml_declaration=True)
+        # In Python 3, val will be bytes
+        if not isinstance(val, str):
+            return str(val, 'utf-8')
+        return val
 
     def add_file_entry(self, path, mimetype=None):
         manifest_namespace = self.namespaces['manifest']
@@ -738,14 +987,69 @@ class Manifest(object):
                                   nsmap={'manifest': manifest_namespace})
         self.root.append(entry_node)
 
+    def remove_file_entry(self, path):
+        manifest_namespace = self.namespaces['manifest']
+        entry = self.root.find('{%s}%s[@{%s}full-path="%s"]' %
+                               (manifest_namespace, 'file-entry',
+                                manifest_namespace, path))
+        if entry is not None:
+            self.root.remove(entry)
+
+
+class Meta(object):
+
+    def __init__(self, content):
+        self.tree = lxml.etree.parse(BytesIO(content))
+        root = self.tree.getroot()
+        self.namespaces = root.nsmap
+        path = '/office:document-meta/office:meta'
+        self.office_meta, = self.tree.xpath(path, namespaces=self.namespaces)
+
+    def set(self, name, value, namespace='meta'):
+        namespace = self.namespaces[namespace]
+        meta = self.office_meta.find('{%s}%s' % (namespace, name))
+        if meta is None:
+            meta = EtreeElement('{%s}%s' % (namespace, name),
+                                nsmap={'meta': namespace})
+            self.office_meta.append(meta)
+        meta.text = value
+
+    def remove(self, name, namespace='meta'):
+        namespace = self.namespaces[namespace]
+        meta = self.office_meta.find('{%s}%s' % (namespace, name))
+        if meta is not None:
+            self.office_meta.remove(meta)
+
+    def __str__(self):
+        now = datetime.datetime.now()
+        self.set('creation-date', now.isoformat())
+        self.set('date', now.isoformat())
+        self.remove('document-statistic')
+        self.set('editing-cycles', '1')
+        self.remove('editing-duration')
+        self.set('generator', 'relatorio/%s' % relatorio.__version__)
+        self.remove('initial-creator')
+        self.remove('print-date')
+        self.remove('printed-by')
+        self.remove('creator', 'dc')
+        self.remove('date', 'dc')
+        val = lxml.etree.tostring(self.tree, encoding='UTF-8',
+                                  xml_declaration=True)
+        # In Python 3, val will be bytes
+        if not isinstance(val, str):
+            return str(val, 'utf-8')
+        return val
+
 
 class OOSerializer:
 
-    def __init__(self, oo_path):
-        self.inzip = zipfile.ZipFile(oo_path)
+    def __init__(self, source):
+        self.inzip = get_zip_file(source)
         self.manifest = Manifest(self.inzip.read(MANIFEST))
-        self.new_oo = StringIO()
-        self.outzip = zipfile.ZipFile(self.new_oo, 'w')
+        self.meta = Meta(self.inzip.read(META))
+        self.new_oo = BytesIO()
+        self.outzip = zipfile.ZipFile(
+            self.new_oo, mode='w', compression=zipfile.ZIP_DEFLATED)
         self.xml_serializer = genshi.output.XMLSerializer()
 
     def __call__(self, stream):
@@ -757,6 +1061,7 @@ class OOSerializer:
             files.setdefault(stream_for, []).append((kind, data, pos))
 
         now = time.localtime()[:6]
+        manifest_info = None
         for f_info in self.inzip.infolist():
             if f_info.filename.startswith('ObjectReplacements'):
                 continue
@@ -767,12 +1072,20 @@ class OOSerializer:
                 new_info = zipfile.ZipInfo(f_info.filename, now)
                 for attr in ('compress_type', 'flag_bits', 'create_system'):
                     setattr(new_info, attr, getattr(f_info, attr))
-                serialized_stream = output_encode(self.xml_serializer(stream))
+                serialized_stream = output_encode(self.xml_serializer(stream),
+                    encoding='utf-8')
                 self.outzip.writestr(new_info, serialized_stream)
             elif f_info.filename == MANIFEST:
-                self.outzip.writestr(f_info, str(self.manifest))
+                manifest_info = f_info
+            elif f_info.filename == META:
+                self.outzip.writestr(f_info, str(self.meta))
+            elif f_info.filename.startswith(THUMBNAILS + '/'):
+                self.manifest.remove_file_entry(f_info.filename)
             else:
                 self.outzip.writestr(f_info, self.inzip.read(f_info.filename))
+        self.manifest.remove_file_entry(THUMBNAILS + '/')
+        if manifest_info:
+            self.outzip.writestr(manifest_info, str(self.manifest))
         self.inzip.close()
         self.outzip.close()
 
